package org.netkernel.module.standard.builtin.mapper;

import java.util.*;

import org.netkernel.container.ILogger;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaBuilder;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.RequestScopeClassLoader;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.StandardModule;
import org.netkernel.module.standard.StandardSpaceMeta;
import org.netkernel.module.standard.builtin.PrivateFilterEndpoint;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.IStandardOverlay;
import org.netkernel.module.standard.endpoint.StandardMonoEndpointImpl;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl.IConfig;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.ISpace;
import org.netkernel.util.SynchronousRequestee;
import org.w3c.dom.*;


public class MapperConfig implements IConfig
{
	private Mapping[] mPublicMappings;
	private Mapping[] mAllMappings;
	private ISpaceElements mElements;
	private TreeMap<String, Mapping> mMappingsById;
	private ISpace mDelegateSpace;
	private ISpace mPrivateMappingsSpace;
	private MapperOverlay mEndpoint;
	private int mEndpointIndex=0;
	
	private static final ISpace[] sNoSpaces = IStandardEndpoint.NO_SPACES;
	
	private static final PairList VALIDATION = new PairList(4);
	static
	{	VALIDATION.put("/*[name()!='config']", "MSG_STD_MAPPER_BAD_ROOT");
		VALIDATION.put("/config/*[name()!='endpoint' and name()!='import']", "MSG_STD_MAPPER_BAD_EP_TAG");
		VALIDATION.put("/config/endpoint/*[name()!='grammar' and name()!='id' and name()!='verbs' and name()!='request' and name()!='private' and name()!='protected' and name()!='header' and "+StandardModule.STANDARD_METADATA_PARAMETERS_VALIDATION+"]", "MSG_STD_MAPPER_BAD_EP_CONTENT");
		VALIDATION.put("/config/endpoint[(count(private)+count(protected))>1]", "MSG_STD_MAPPER_BAD_PRIVACY");
		VALIDATION.put("/config/endpoint[not(request)]", "MSG_STD_MAPPER_MISSING_REQUEST");
		VALIDATION.put("/config/endpoint/request/verb", "MSG_STD_MAPPER_REQ_NO_VERB");
		VALIDATION.put("/config/endpoint/request/argument[@name='primary']", "MSG_STD_MAPPER_REQ_NO_PRIMARY");
		VALIDATION.put("/config/endpoint/header[not(@name)]", "MSG_STD_MAPPER_HEADER_NO_NAME");
	}	
	
	public MapperConfig(Document aConfig, ISpace aDelegateSpace, ISpaceMeta aDelegateMeta, INKFRequestContext aContext, MapperOverlay aEndpoint ) throws Exception
	{
		mDelegateSpace=aDelegateSpace;
		mEndpoint=aEndpoint;
		mMappingsById = new TreeMap<String, Mapping>(StandardSpaceMeta.sComparator);

		List<Mapping> publicMappings = new ArrayList<Mapping>();
		List<Mapping> privateMappings = new ArrayList<Mapping>();
		List<Mapping> allMappings = new ArrayList<Mapping>();
		Map<String,IMetaRepresentation> metaCache = new HashMap<String,IMetaRepresentation>();
		recurseGetMappings(publicMappings,privateMappings,allMappings,aConfig,aContext,aDelegateSpace,aEndpoint,metaCache);
		
		
		mAllMappings = new Mapping[allMappings.size()];
		allMappings.toArray(mAllMappings);
		mPublicMappings = new Mapping[publicMappings.size()];
		publicMappings.toArray(mPublicMappings);
		Mapping[] privateMappingArray = new Mapping[privateMappings.size()];
		privateMappings.toArray(privateMappingArray);
		
		//build space elements
		ISpaceElements result;
		TupleList parts = new TupleList(3,mPublicMappings.length);
		for (int i=0; i<mPublicMappings.length; i++)
		{	Mapping m=mPublicMappings[i];
			String id = m.getId();
			Mapping existing=mMappingsById.put(id, m);
			if (existing!=null)
			{	aContext.logFormatted(INKFLocale.LEVEL_WARNING, "MSG_STD_MAPPER_DUPLICATE_ID", id, aEndpoint.getOwningSpace().getOwningModule().getMeta().getName());
			}
			
			ISpace delegateSpace=null;
			if (m.isTransparent())
			{	delegateSpace=aDelegateSpace;
			}
			
			parts.put(new SimpleIdentifierImpl(id), delegateSpace,aEndpoint);
		}
		mElements = new SpaceElementsImpl(parts);
		
		//build space for private mappings
		if (privateMappings.size()>0)
		{	MapperPrivateSpace mps = new MapperPrivateSpace(privateMappingArray,mEndpoint);
			mps.setName("MapperPrivateSpace");
			mps.onCommissionSpace(aContext.getKernelContext().getKernel());
			mPrivateMappingsSpace = mps;
		}
	}
	
	private void recurseGetMappings(List<Mapping> aPublicMappings, List<Mapping> aPrivateMappings, List<Mapping> aAllMappings, Document aConfig, INKFRequestContext aContext, ISpace aDelegateSpace, MapperOverlay aEndpoint, Map<String,IMetaRepresentation> aMetaCache) throws Exception
	{
		PairList validation=FastSchematron.validate(aConfig.getDocumentElement(), VALIDATION);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,aContext.getKernelContext().getKernel().getLogger());			
			throw Utils.createFormattedException("EX_STD_MAPPER_CONFIG", "MSG_RAW", aContext.getKernelContext().getKernel().getLogger(), m);
		}
		
		XMLReadable r = new XMLReadable(aConfig);
		List nodes=r.getNodes("/config/*");
		for (int i=0; i<nodes.size(); i++)
		{	Node n=(Node)nodes.get(i);
			String nodeName=XMLUtils.pleaseDOMGiveMeTheNameOf(n);
			if (nodeName.equals("endpoint"))
			{	try
				{	Mapping m = new Mapping(n,aContext,aDelegateSpace,aEndpoint,aMetaCache);
					if (m.isPrivate())
					{	aPrivateMappings.add(m);
					}
					else
					{	aPublicMappings.add(m);
					}
					aAllMappings.add(m);
				}
				catch (Exception e)
				{	//tolerant of failure of specific mappings
					String spaceName=aEndpoint.getOwningSpace().toString();
					String path=XMLUtils.getPathFor(n);
					String cause;
					if (e instanceof INetKernelException)
					{	cause = ((INetKernelException)e).getDeepestId()+": "+((INetKernelException)e).getDeepestMessage();
					}
					else
					{	cause = e.getClass().getName()+": "+e.getMessage();
					}
					aContext.logFormatted(INKFLocale.LEVEL_WARNING, "MSG_STD_MAPPER_CONSTRUCT_MAPPING",spaceName,path,cause);
				}
			}
			else if (nodeName.equals("import"))
			{	String id=XMLUtils.getText(n).trim();
				try
				{	Document d=Utils.sourceAsDOMAvoidingTransrept(aContext, id);
					recurseGetMappings(aPublicMappings,aPrivateMappings,aAllMappings,d,aContext,aDelegateSpace,aEndpoint,aMetaCache);
				}
				catch (Exception e)
				{	//tolerant of failure of import
					aContext.logRaw(INKFLocale.LEVEL_WARNING, Utils.throwableToString(e));
				}
			}
		}
	}
	
	public IMetaRepresentation getElementMeta(String aElementId)
	{	String absoluteElementId = getAbsoluteIdentifier(aElementId);
		Mapping m=mMappingsById.get(absoluteElementId);
		IMetaRepresentation result=null;
		if (m!=null)
		{	result=m.getMeta();
		}
		return result;
	}
	
	private String getAbsoluteIdentifier(String aIdentifier)
	{	String result=null;
		Mapping mapping=mMappingsById.get(aIdentifier);
		if (mapping!=null)
		{	result=aIdentifier;
		}
		else
		{	SortedMap<String,Mapping> m=mMappingsById.tailMap(aIdentifier);
			if (!m.isEmpty())
			{	String possibleMatch = m.firstKey();
				if (possibleMatch.endsWith(aIdentifier))
				{	result=possibleMatch;
				}
			}
		}
		return result;
	}	

	public ISpaceElements getElements(INKFRequestContext aContext) throws Exception
	{	return mElements;
	}
	

	public ISpace[] getSpaces(INKFRequestContext context) throws Exception
	{	ISpace[] result;
	 	if (mPrivateMappingsSpace!=null)
	 	{	result = new ISpace[] {mPrivateMappingsSpace};
	 	}
	 	else
	 	{	result = sNoSpaces;
	 	}
		return result;
	}

	public String getResolvedElement(INKFResolutionContext aContext) throws Exception
	{	String result=null;
		INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
		for (Mapping mapping : mPublicMappings)
		{	IEndpointMeta meta = mapping.getMeta();
			if (MetadataUtils.match((IEndpointMeta)meta, toResolve, aContext, mapping.getId()))
			{	result = mapping.getId();
				break;
			}
		}
		return result;
	}

	public void onRequest(String aElementId, INKFRequestContext aContext) throws Exception
	{	Mapping mapping = mMappingsById.get(aElementId);
		mapping.onRequest(aContext);
	}
	
	private static INetKernelException generateConfigException(String aMsg, ILogger aLogger, Node aLocation, Object aOptionalParam)
	{	return Utils.createFormattedException("EX_STD_MAPPER_CONFIG", aMsg, aLogger, XMLUtils.getPathFor(aLocation), aOptionalParam);
	}
	
	
	private int getEndpointIndex()
	{	return mEndpointIndex++;
	}
	
	public IHDSNode getConfig()
	{
		HDSBuilder b=new HDSBuilder();
		b.pushNode("config");
		for (Mapping m : mAllMappings)
		{	
			b.pushNode("endpoint");
			b.addNode("id", m.mId);
			b.addNode("name",m.mMeta.getName());
			IRequestResponseFields headers=m.mMeta.getAdditionalFields();
			String visibility=m.mPrivate?"private":(Boolean.TRUE.equals(headers.getValue(PrivateFilterEndpoint.KEY_PRIVATE))?"protected":"public");
			b.addNode("visibility",visibility);
			if (m.mRequestBuilder!=null)
			{	m.mRequestBuilder.writeHDS(b);
			}
			b.popNode();
		}
		return b.getRoot();
	}
	
	
	public class Mapping
	{
		private String mId;
		private IEndpointMeta mMeta;
		private RequestBuilder mRequestBuilder;
		private boolean mPrivate;
		private PairList mHeaders;
		
		private Mapping(Node aNode, INKFRequestContext aContext, ISpace aHostSpace, IEndpoint aHostEndpoint, Map<String,IMetaRepresentation> aMetaCache) throws Exception
		{	XMLReadable r = new XMLReadable(aNode);
			ILogger logger=aContext.getKernelContext().getKernel().getLogger();
			
			IEndpointMeta targetMeta=null;
			String targetId=null;
			String identifier=r.getText("request/identifier");
			final String META_SCHEME="meta:";
			if (identifier.startsWith(META_SCHEME))
			{	targetId=identifier.substring(META_SCHEME.length());
				
				IMetaRepresentation meta=aMetaCache.get(targetId);
				if (meta==null)
				{	meta = aContext.meta(targetId);
					aMetaCache.put(targetId,meta);
				}
				if (meta instanceof IEndpointMeta)
				{	targetMeta=(IEndpointMeta)meta;
				}
				else
				{	throw generateConfigException("MSG_STD_MAPPER_TARGET_NOT_EP", logger, aNode, meta.getClass().getName());
				}
			}
			
			Element grammarNode=XMLUtils.getChildElementNamed(aNode, "grammar");
			boolean needRequestBuilder = targetId==null || r.getNodes("request/argument").size()>0 || grammarNode!=null;
			boolean inheritTargetMeta=targetId!=null && r.getNodes("request/argument").size()==0;
			
			//determine id
			String id=r.getText("id");
			if (id.length()==0)
			{	if (inheritTargetMeta)
				{	id=targetId;
				}
				else
				{	int mappingIndex=getEndpointIndex();
					id=Utils.generateUniqueEndpointIdentifier(aHostSpace, aHostEndpoint, Integer.toString(mappingIndex));
				}
			}
			mId=id;
			
			//determine if private
			mPrivate=r.getNodes("private").size()>0;

			//determine name
			//boolean hasAutoName=false;
			String name=r.getText("name");
			if (name.length()==0)
			{	if (r.getText("id").length()!=0)
				{	name=id;
				}
			}
			
			//determine description
			String desc=r.getText("description");
			if (desc.length()==0 && inheritTargetMeta)
			{	desc=targetMeta.getDescription();
			}

			//determine icon
			String iconString=r.getText("icon");
			IIdentifier icon=null;
			if (iconString.length()>0)
			{	icon = new SimpleIdentifierImpl(iconString);
			}
			else if (inheritTargetMeta)
			{	icon=targetMeta.getIcon();
			}
			
			//determine doc
			String hrdString=r.getText("doc");
			IIdentifier hrd=null;
			if (hrdString.length()>0)
			{	hrd = new SimpleIdentifierImpl(hrdString);
			}
			else if (inheritTargetMeta)
			{	hrd=targetMeta.getHumanReadableDocumentation();
			}
			
			EndpointMetaBuilder builder = new EndpointMetaBuilder();
			builder.setDescription(desc);
			builder.setIcon(icon);
			builder.setHumanReadableDocumentation(hrd);
			
			Node userMetaNode=r.getFirstNode(StandardMonoEndpointImpl.PARAM_META);
			if (userMetaNode!=null)
			{	builder.addField(StandardMonoEndpointImpl.PARAM_META, userMetaNode);
			}
			
			
			//grammar
			if (grammarNode!=null)
			{	MetadataUtils.parseGrammar(grammarNode, builder, logger);
				if (name.length()==0)
				{	name=builder.getGrammar(logger).toString();
				}
			}
			else
			{	//only valid to have no grammar if we don't need to parse request
				if (needRequestBuilder)
				{	throw generateConfigException("MSG_STD_MAPPER_ARGS_WITHOUT_GRAMMAR", logger, aNode, null);
				}
				else
				{	//inherit interface from target
					IIdentifierGrammar targetGrammar=targetMeta.getIdentifierGrammar();
					builder.setGrammar(targetGrammar);
				}
			}
			
			if (name.length()==0)
			{	//hasAutoName=true;
				if (inheritTargetMeta && r.getText("id").length()==0)
				{	name=targetMeta.getName();
				}
				else
				{	name=id;
				}
			}
			builder.setName(name);
			
			//override verbs
			String verbsString=r.getText("verbs");
			if (verbsString.length()>0)
			{	try
				{	int verbs=Utils.parseVerbList(verbsString);
					builder.setVerbs(verbs);
				}
				catch (IllegalArgumentException e)
				{	throw generateConfigException("MSG_STD_MAPPER_BAD_VERB", logger, aNode, verbsString);
				}
			}
			else
			{	if (targetMeta!=null)
				{	builder.setVerbs(targetMeta.getSupportedVerbs());
				}
				else
				{	//default to SOURCE only
					builder.setVerbs(IRequest.VERB_SOURCE);
				}
			}
			
			//set private meta if marked protected
			if (r.getNodes("protected").size()>0)
			{	builder.addField(PrivateFilterEndpoint.KEY_PRIVATE, Boolean.TRUE);
			}
			
			mMeta=builder.getMeta(logger);
			
			if (needRequestBuilder)
			{	Node request=r.getFirstNode("request");
				mRequestBuilder = new RequestBuilder(request,logger);
				mRequestBuilder.setPassthroughRequestedRepresentation(true);
			}
			
			List<Node> headerNodes=r.getNodes("header");
			if (headerNodes.size()>0)
			{	mHeaders=new PairList(headerNodes.size());
				for (Node headerNode : headerNodes)
				{	String headerName=((Element)headerNode).getAttribute("name");
				
					Object headerValue;
					Element literalNode=XMLUtils.getFirstChildElement(headerNode);
					if (literalNode==null)
					{	headerValue=XMLUtils.getText(headerNode);
					}
					else if (XMLUtils.getNextSiblingElement(literalNode)==null)
					{	ClassLoader cl=new RequestScopeClassLoader(aContext.getKernelContext().getRequestScope());
						headerValue=RequestBuilder.parseLiteral(literalNode, cl, aContext)[0];
					}
					else
					{	throw generateConfigException("MSG_STD_MAPPER_HEADER_MULTILITERAL", logger, aNode, headerName);
					}				
					mHeaders.put(headerName, headerValue);
				}
			}
		}
		
		public String getId()
		{	return mId;
		}
		
		public IEndpointMeta getMeta()
		{	return mMeta;
		}
		
		public boolean isPrivate()
		{	return mPrivate;
		}
		
		public boolean isTransparent()
		{	return mRequestBuilder==null;
		}
		
		public void onRequest(INKFRequestContext aContext) throws Exception
		{
			INKFRequest req;
			if (mRequestBuilder==null)
			{	//transparent relay of request
				req = aContext.getThisRequest().getIssuableClone();
			}
			else
			{	//build sub-request
				req = mRequestBuilder.buildRequest(aContext, null, null);
				req.setVerb(aContext.getThisRequest().getVerb());
				INKFResponseReadOnly primary=aContext.getThisRequest().getPrimaryAsResponse();
				if (primary!=null)
				{	req.addPrimaryArgumentFromResponse(primary);
				}
			}
			
			if (mPrivateMappingsSpace!=null)
			{	// construct a scope with private mappings and then delegated space
				IRequestScopeLevel scope=aContext.getKernelContext().getThisKernelRequest().getRequestScope();
				scope=RequestScopeLevelImpl.createNonDurableScopeLevel(mPrivateMappingsSpace, scope);
				scope=RequestScopeLevelImpl.createNonDurableScopeLevel(mDelegateSpace, scope);
				req.setRequestScope(scope);
			}
			
			IEndpoint ep=mElements.getPhysicalEndpoint(0);
			//req.setHeader(IStandardOverlay.HEADER_REQUESTOR_OVERLAY, ep); //tells overlay to detect request going outside this overlay
			if(!aContext.getKernelContext().isThisRequestAsync())
			{	//Synchronous
				//System.out.println("SYNC MAPPING");
				INKFResponseReadOnly resp=aContext.issueRequestForResponse(req);
				INKFResponse respOut=aContext.createResponseFrom(resp);
				if (mHeaders!=null)
				{	for (int i=0; i<mHeaders.size(); i++)
					{	respOut.setHeader((String)mHeaders.getValue1(i), mHeaders.getValue2(i));
					}
				}
			}
			else
			{	//Asynchronous
				//System.out.println("ASYNC MAPPING!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
				aContext.issueAsyncRequest(req).setListener(new AsyncMappingResponseHandler());
				aContext.setNoResponse();
				
			}
		}
		
		public class AsyncMappingResponseHandler implements INKFAsyncRequestListener
		{
			@Override
			public void receiveResponse(INKFResponseReadOnly aResponse,
					INKFRequestContext aContext) throws Exception
			{	INKFResponse respOut=aContext.createResponseFrom(aResponse);
				if (mHeaders!=null)
				{	for (int i=0; i<mHeaders.size(); i++)
					{	respOut.setHeader((String)mHeaders.getValue1(i), mHeaders.getValue2(i));
					}
				}
			}

			@Override
			public void receiveException(NKFException aException,
					INKFRequest aRequest, INKFRequestContext aContext)
					throws Exception
			{	aContext.createResponseFrom(aException);
			}
		}
	}
}
